"""A collection of common stream operations."""
from ._buffers import Window, Group, Delay
from ._combiners import Chain, Merge, Zip, ZipLatest
from ._filters import Skip, Filter, Changes, RateLimit
from ._producers import Repeat
from ._refiners import Map



""" Abstract Base Classes for building pipeline sections. """
from abc import ABC, abstractmethod
from typing import Any, AsyncIterable, Awaitable, Callable, Iterable, Optional

class Section(ABC):
    """Each pipeline section takes inputs from an async iterable, processes it and sends it to an
    output.

    A section must implement the :meth:`pump` abstract method, which will be scheduled to run as a task
    by the pipeline. The pump method serves as an underlying machinery for pulling and pushing
    data through the section.

    Each subclass of section must implement a way to process each received item. This should be
    done by either implementing an asynchronous or synchronous refine method. Slurry defines two
    standard apis for implementing a refine method, an asynchronous one as defined by
    :class:`AsyncSection`, and a synchronous as defined by :class:`SyncSection`. The refine api
    is not strictly a requirement, but a convention. Any class that implements the ``Section``
    api is a valid pipeline section.
    """

    @abstractmethod
    async def pump(self, input: Optional[AsyncIterable[Any]], output: Callable[[Any], Awaitable[None]]):
        """The pump method contains the machinery that takes input from previous sections, or
        any asynchronous iterable, processes it and pushes it to the output.

        .. note::
            The receiving end of the output can be closed by the pipeline or by the downstream
            section at any time. If you try to send an item to an output that has a closed receiver,
            a ``BrokenResourceError`` will be raised. The pipeline knows about this and is prepared
            to handle it for you, but if you need to do some kind of cleanup, like closing network
            connections for instance, you may want to handle this exception yourself.

        .. note::
            If this section is the first section of a pipeline, the input will be ``None``. In this
            case, the section is expected to produce output independently.

        :param input: The input data feed. Will be ``None`` for the first ``Section``, as the first
            ``Section`` is expected to supply it's own input.
        :type input: Optional[AsyncIterable[Any]]
        :param output: An awaitable callable used to send output.
        :type output: Callable[[Any], Awaitable[None]]
        """

class AsyncSection(Section):
    """AsyncSection defines an abc for sections that are designed to run in an async event loop."""

    @abstractmethod
    async def refine(self, input: Optional[AsyncIterable[Any]], output: Callable[[Any], Awaitable[None]]):
        """The async section refine method must contain the logic that iterates the input, processes
        the indidual items, and feeds results to the output.

        By default, the pipeline tries to manage the input and output resource lifetime. Normally
        you don't have to worry about closing the input and output after use. The exception is, if
        your custom section adds additional input sources, or provides it's own input. In this case
        the section must take care of closing the input after use.

        :param input: The input data feed. Will be ``None`` for the first ``Section``, as the first
            ``Section`` is expected to supply it's own input.
        :type input: Optional[AsyncIterable[Any]]
        :param output: An awaitable callable used to send output.
        :type output: Callable[[Any], Awaitable[None]]
        """

class SyncSection(Section):
    """Sync defines an abc for sections that runs synchronous refiners. Implementations
    of ``SyncSection`` should be carefully designed so as to not block the underlying
    asynchronous pipeline event loop.

    Async sections and synchronous sections can be freely mixed and matched in the pipeline.
    """

    @abstractmethod
    def refine(self, input: Optional[Iterable[Any]], output: Callable[[Any], None]):
        """The ``SyncSection`` refine method is intended to run normal synchronous python
        code, including code that can block for IO for an any amount of time. Implementations
        of ``SyncSection`` should take care to design a pump method in such a way, that blocking
        happens transparently to the parent async event loop.

        The refine method is designed to have an api that is as close to the async api as
        possible. The input is a synchronous iterable instead of an async iterable, and the
        output is a synchronous callable, similar to a ``Queue.put`` method.

        .. note::
            Slurry includes two implementations of ``SyncSection``.
            :class:`slurry.environments.ThreadSection`, which runs the refine function in
            a background thread, and :class:`slurry.environments.ProcessSection`
            which spawns an independent process that runs the refine method.

        :param input: The input data feed. Like with ordinary sections, this can be ``None`` if
            ``SyncSection`` is the first section in the pipeline.
        :type input: Optional[Iterable[Any]]
        :param output: The callable used to send output.
        :type output: Callable[[Any], None]
        """



"""The Trio environment implements ``TrioSection``, which is a Trio-native 
:class:`AsyncSection <slurry.sections.abc.AsyncSection>`."""
from typing import Any, AsyncIterable, Awaitable, Callable, Optional

from ..sections.abc import AsyncSection

class TrioSection(AsyncSection):
    """Since Trio is the native Slurry event loop, this environment is simple to implement.
    The pump method does not need to do anything special to bridge the input and output. It
    simply delegates directly to the refine method, as the api is identical."""
    async def pump(self, input: Optional[AsyncIterable[Any]], output: Callable[[Any], Awaitable[None]]):
        """Calls refine."""
        await self.refine(input, output)



"""Pipeline sections for combining multiple inputs into a single output."""
import builtins
import itertools
from typing import Any, AsyncIterable, Sequence

import trio
from async_generator import aclosing

from ..environments import TrioSection
from .weld import weld

class Chain(TrioSection):
    """Chains input from one or more sources. Any valid ``PipelineSection`` is an allowed source.

    Outputs items from each source in turn, until it is exhausted. If a source never reaches the
    end, remaining sources will not be iterated.

    Chain can be placed as a middle section and will chain the input of the previous section.

    .. Note::
        By default, the input is added as the first source. If the input is added last instead
        of first, it will cause backpressure to be applied upstream.

    :param sources: One or more ``PipelineSection`` that will be chained together.
    :type sources: Sequence[PipelineSection]
    :param place_input: Iteration priority of the pipeline input source. Options:
        ``'first'`` (default) \\| ``'last'``.
    :type place_input: string
    """
    def __init__(self, *sources: Sequence["PipelineSection"], place_input='first'):
        super().__init__()
        self.sources = sources
        self.place_input = _validate_place_input(place_input)

    async def refine(self, input, output):
        if input:
            if self.place_input == 'last':
                sources = (*self.sources, input)
            else:
                sources = (input, *self.sources)
        else:
            sources = self.sources
        async with trio.open_nursery() as nursery:
            for source in sources:
                async with aclosing(weld(nursery, source)) as agen:
                    async for item in agen:
                        await output(item)

class Merge(TrioSection):
    """Merges input from multiple sources. Any valid ``PipelineSection`` is an allowed source.

    Sources are iterated in parallel and items are send from each source, as soon as
    they become available.

    If Merge is used as a middle section, the input will be added to the sources.

    Sources can be pipeline sections, which will be treated as first sections, with
    no input. Merge will take care of running the pump task for these sections.

    :param sources: One or more async iterables or sections who's contents will be merged.
    :type sources: Sequence[PipelineSection]
    """
    def __init__(self, *sources: Sequence["PipelineSection"]):
        super().__init__()
        self.sources = sources

    async def refine(self, input, output):
        async with trio.open_nursery() as nursery:

            async def pull_task(source):
                async with aclosing(weld(nursery, source)) as aiter:
                    async for item in aiter:
                        await output(item)

            if input:
                nursery.start_soon(pull_task, input)

            for source in self.sources:
                nursery.start_soon(pull_task, source)

class Zip(TrioSection):
    """Zips the input from multiple sources. Any valid ``PipelineSection`` is an allowed source.

    Sources are iterated in parallel and as soon as all sources have an item available, those
    items are output as a tuple.

    Zip can be used as a middle section, and the pipeline input will be added to the sources.

    .. Note::
        If sources are out of sync, the fastest source will have to wait for the slowest, which
        will cause backpressure.

    :param sources:  One or more ``PipelineSection``, who's contents will be zipped.
    :type sources: Sequence[PipelineSection]
    :param place_input:  Position of the pipeline input source in the output tuple. Options:
        ``'first'`` (default) \\| ``'last'``.
    :type place_input: string
    """
    def __init__(self, *sources: Sequence["PipelineSection"], place_input='first'):
        super().__init__()
        self.sources = sources
        self.place_input = _validate_place_input(place_input)

    async def refine(self, input, output):
        if input:
            if self.place_input == 'last':
                sources = (*self.sources, input)
            else:
                sources = (input, *self.sources)
        else:
            sources = self.sources

        async with trio.open_nursery() as weld_nursery:
            sources = [weld(weld_nursery, source) for source in sources]

            async def pull_task(source, index, results: list):
                try:
                    results.append((index, await source.__anext__()))
                except StopAsyncIteration:
                    weld_nursery.cancel_scope.cancel()

            while True:
                results = []
                async with trio.open_nursery() as pull_nursery:
                    for i, source in builtins.enumerate(sources):
                        pull_nursery.start_soon(pull_task, source, i, results)
                await output(tuple(result for i, result in sorted(results, key=lambda packet: packet[0])))

        for source in sources:
            await source.aclose()

class ZipLatest(TrioSection):
    """Zips input from multiple sources and outputs a result on every received item. Any valid
    ``PipelineSection`` is an allowed source.

    Sources are iterated in parallel and a tuple is output each time a result is ready
    on any source. The tuple values will be the last received value from each source.

    Using the monitor argument, one or more asynchronous sequences can be added with the property
    that they will not trigger an output by themselves. Their latest value will be stored and
    added to the output value, but will only be output if a new item arrives at one of the main
    sources.

    ZipLatest can be used as a middle section, in which case the upstream pipeline is
    added as an input.

    .. Note::
        If any single source is exchausted, all remaining sources will be forcibly closed, and
        the pipeline will stop.

    :param sources: One or more async iterables that will be zipped together.
    :type sources: Sequence[AsyncIterable[Any]]
    :param partial: If ``True`` (default) output will be sent as soon as the first input arrives.
        Otherwise, all main sources must send at least one item, before an output is generated.
    :type partial: bool
    :param default: If the parameter ``partial`` is ``True``, this value is used as the
        default value to output, until an input has arrived on a source. Defaults to ``None``.
    :type default: Any
    :param monitor: Additional asynchronous sequences to monitor.
    :type monitor: Optional[Union[AsyncIterable[Any], Sequence[AsyncIterable[Any]]]]
    :param place_input: Position of the pipeline input source in the output tuple. Options:
        ``'first'`` (default)|``'last'``
    :type place_input: string
    :param monitor_input: Input is used as a monitored stream instead of a main source.
        Defaults to ``False``
    :type monitor_input: bool
    """
    def __init__(self, *sources: Sequence["PipelineSection"],
                 partial=True,
                 default=None,
                 monitor=(),
                 place_input='first',
                 monitor_input=False):
        super().__init__()
        self.sources = sources
        self.partial = partial
        self.default = default
        self.monitor = monitor
        self.place_input = _validate_place_input(place_input)
        self.monitor_input = monitor_input

    async def refine(self, input, output):
        sources = self.sources
        try:
            iter(self.monitor)
            monitor = self.monitor
        except TypeError:
            monitor = (self.monitor,)

        if input is not None:
            if self.monitor_input:
                if self.place_input == 'last':
                    monitor = (*monitor, input)
                else:
                    monitor = (input, *monitor)
            else:
                if self.place_input == 'last':
                    sources = (*sources, input)
                else:
                    sources = (input, *sources)

        results = [self.default for _ in itertools.chain(sources, monitor)]
        ready = [False for _ in results]

        async with trio.open_nursery() as nursery:

            async def pull_task(index, source, monitor=False):
                async with aclosing(weld(nursery, source)) as aiter:
                    async for item in aiter:
                        results[index] = item
                        ready[index] = True
                        if not monitor and (self.partial or False not in ready):
                            await output(tuple(results))
                nursery.cancel_scope.cancel()

            for i, source in builtins.enumerate(sources):
                nursery.start_soon(pull_task, i, source)
            for i, source in builtins.enumerate(monitor):
                nursery.start_soon(pull_task, i + len(sources), source, True)

def _validate_place_input(place_input):
    if isinstance(place_input, str):
        if place_input not in ['first', 'last']:
            raise ValueError(f'Invalid place_input argument: {place_input}')
    else:
        raise TypeError('place_input argument has invalid type.')
    return place_input



"""The weld module implements the ``weld`` function that connects ``PipelineSection`` objects 
together and returns the async iterable output.

A ``PipelineSection`` is any object that is valid input to the ``weld`` function. This currently
includes the following types:

* ``AsyncIterable[Any]`` - Async iterables are valid only as the very first ``PipelineSection``. Subsequent
  sections will use this async iterable as input source. Placing an ``AsyncIterable`` into the middle of
  of a sequence of pipeline sections, will cause a ``ValueError``.
* Any :class:`Section` abc subclass.
* ``Tuple[PipelineSection, ...]`` - Pipeline sections can be nested to any level by supplying a tuple
  containing one or more pipeline section-compatible objects. Output from upstream sections are
  automatically used as input to the nested sequence of pipeline sections.

The main :class:`Pipeline <slurry.Pipeline>` class uses the ``weld`` function to compose the sequence of
``PipelineSection`` objeects (see below) and return an output. Similarly, the
:ref:`combiner sections <Combining multiple inputs>` use the ``weld`` function to support defining
sub-pipelines as input sources, using the tuple notation."""

from typing import Any, AsyncIterable, Optional, Sequence

import trio

from .abc import Section

def weld(nursery, *sections: Sequence["PipelineSection"]) -> AsyncIterable[Any]:
    """
    Connects the individual parts of a sequence of pipeline sections together and starts pumps for
    individual Sections. It returns an async iterable which yields results of the sequence.

    :param nursery: The nursery that runs individual pipeline section pumps.
    :type nursery: :class:`trio.Nursery`
    :param \\*sections: A sequence of pipeline sections.
    :type \\*sections: Sequence[PipelineSection]
    """

    async def pump(section, input: Optional[AsyncIterable[Any]], output: trio.MemorySendChannel):
        try:
            await section.pump(input, output.send)
        except trio.BrokenResourceError:
            pass
        if input:
            await input.aclose()
        await output.aclose()

    section_input = None
    output = None
    for section in sections:
        if isinstance(section, Section):
            section_output, output = trio.open_memory_channel(0)
            nursery.start_soon(pump, section, section_input, section_output)
        elif isinstance(section, tuple):
            if section_input:
                output = weld(nursery, section_input, *section)
            else:
                output = weld(nursery, *section)
        else:
            if output:
                raise ValueError('Invalid pipeline section.', section)
            output = section
        section_input = output

    return output



"""Pipeline output tap."""
import trio

class Tap:
    """The tap class is responsible for transmitting the output of the pipeline to consumers.
    It implements an asynchronous send function that is launched as a task from the pipeline for
    each item processed. Each tap can be individually configured with send timeouts and retry
    attempts.

    .. Note::
        This class should not be instantiated by client applications. Create a tap by calling
        :meth:`slurry.pipeline.Pipeline.tap`.

    :param send_channel: The output to which items are sent.
    :type send_channel: trio.MemorySendChannel
    :param timeout: Seconds to wait for receiver to respond.
    :type timeout: float
    :param retrys: Number of times to reattempt a send that timed out.
    :type retrys: int
    """
    def __init__(self, send_channel, timeout, retrys):
        self.send_channel = send_channel
        self.timeout = timeout
        self.retrys = retrys
        self.closed = False

    async def send(self, item):
        """Handles the transmission of a single item from the pipeline.

        Each send operation is run as a task, so that in case of multiple consumers, a stuck
        consumer won't block the entire send loop.

        :param item: The item to send.
        :type item: Any
        """
        for _ in range(self.retrys + 1):
            with trio.move_on_after(self.timeout):
                try:
                    await self.send_channel.send(item)
                except trio.BrokenResourceError:
                    self.closed = True
                return
            await trio.sleep(0)
        raise trio.BusyResourceError('Unable to send item.')



"""Implements a section that runs in an independent python proces."""

from multiprocessing import Process, SimpleQueue
from typing import Any, AsyncIterable, Awaitable, Callable

import trio

from ..sections.abc import SyncSection

class ProcessSection(SyncSection):
    """ProcessSection defines a section interface with a synchronous
    :meth:`refine <slurry.sections.abc.SyncSection.refine>` method that
    runs in a separate process. Slurry makes use of the python
    `multiprocessing <https://docs.python.org/3/library/multiprocessing.html>`_ module
    to spawn the process.

    .. note::
        ``ProcessSection`` implementations must be `pickleable
        <https://docs.python.org/3/library/pickle.html#what-can-be-pickled-and-unpickled>`_.
    """

    async def pump(self, input: AsyncIterable[Any], output: Callable[[Any], Awaitable[None]]):
        """
        The ``ProcessSection`` pump method works similar to the threaded version, however
        since communication between processes is not as simple as it is between threads,
        that are directly able to share memory with each other, there are some restrictions
        to be aware of.

        * Data that is to be sent to the input or transmitted on the output must be `pickleable
          <https://docs.python.org/3/library/pickle.html#what-can-be-pickled-and-unpickled>`_.
        * Since ``ProcessSection`` uses unbounded queues to transfer data behind the scenes, they
          are unable to provide or receive backpressure.
        """
        if input:
            input_queue = SimpleQueue()
        else:
            input_queue = None
        output_queue = SimpleQueue()
        process = Process(target=self._process_run_target,
                          args=(input_queue, output_queue))

        async def sender():
            async for item in input:
                await trio.to_thread.run_sync(input_queue.put, (item,))
            await trio.to_thread.run_sync(input_queue.put, ())

        async with trio.open_nursery() as nursery:
            if input:
                nursery.start_soon(sender)
            process.start()
            while True:
                wrapped_item = await trio.to_thread.run_sync(output_queue.get)
                if wrapped_item == ():
                    break
                await output(wrapped_item[0])
            nursery.cancel_scope.cancel()

    def _process_run_target(self, input_queue: SimpleQueue, output_queue: SimpleQueue):
        if input_queue:
            input = (wrapped_item[0] for wrapped_item in iter(input_queue.get, ()))
        else:
            input = None
        self.refine(input, lambda item: output_queue.put((item,)))
        output_queue.put(())



"""Sections for transforming an input into a different output."""
from typing import Any, AsyncIterable, Optional

from async_generator import aclosing

from ..environments import TrioSection

class Map(TrioSection):
    """Maps over an asynchronous sequence.

    Map can be used as a starting section, if a source is provided.

    :param func: Mapping function.
    :type func: Callable[[Any], Any]
    :param source: Source if used as a starting section.
    :type source: Optional[AsyncIterable[Any]]
    """
    def __init__(self, func, source: Optional[AsyncIterable[Any]] = None):
        self.func = func
        self.source = source

    async def refine(self, input, output):
        if input:
            source = input
        elif self.source:
            source = self.source
        else:
            raise RuntimeError('No input provided.')

        async with aclosing(source) as aiter:
            async for item in aiter:
                await output(self.func(item))



"""Pipeline sections that filters the incoming items."""
from typing import Any, AsyncIterable, Callable, Hashable, Optional, Union

from async_generator import aclosing
import trio

from ..environments import TrioSection

class Skip(TrioSection):
    """Skips the first ``count`` items in an asynchronous sequence.

    Skip can be used as a starting section if a source is given.

    :param count: Number of items to skip
    :type count: int
    :param source: Input source if starting section.
    :type source: Optional[AsyncIterable[Any]]
    """
    def __init__(self, count: int, source: Optional[AsyncIterable[Any]] = None):
        super().__init__()
        self.count = count
        self.source = source

    async def refine(self, input, output):
        if input:
            source = input
        elif self.source:
            source = self.source
        else:
            raise RuntimeError('No input provided.')

        async with aclosing(source) as aiter:
            try:
                for _ in range(self.count):
                    await aiter.__anext__()
            except StopAsyncIteration:
                return
            async for item in aiter:
                await output(item)

class Filter(TrioSection):
    """Outputs items that passes a filter function.

    The filter function must take an item. If the return value evaluates as true, the item is sent,
    otherwise the item is discarded.

    Filter can be used as a starting section, if a source is provided.

    :param func: Matching function.
    :type func: Callable[[Any], bool]
    :param source: Source if used as a starting section.
    :type source: Optional[AsyncIterable[Any]]
    """
    def __init__(self, func, source: Optional[AsyncIterable[Any]] = None):
        super().__init__()
        self.func = func
        self.source = source

    async def refine(self, input, output):
        if input:
            source = input
        elif self.source:
            source = self.source
        else:
            raise RuntimeError('No input provided.')

        async with aclosing(source) as aiter:
            async for item in aiter:
                if self.func(item):
                    await output(item)

class Changes(TrioSection):
    """Outputs items that are different from the last item output.

    The generator stores a reference to the last outputted item. Whenever a new item arrives, it is
    compared to the last outputted item. If they are equal, the new item is discarded. If not,
    the new item is output and becomes the new reference. The first item received is always
    output.

    Changes can be used as a starting section, if a source is provided.

    .. Note::
        Items are compared using the != operator.

    :param source: Source if used as a starting section.
    :type source: Optional[AsyncIterable[Any]]
    """
    def __init__(self, source: Optional[AsyncIterable[Any]] = None):
        super().__init__()
        self.source = source

    async def refine(self, input, output):
        if input:
            source = input
        elif self.source:
            source = self.source
        else:
            raise RuntimeError('No input provided.')

        token = object()
        last = token
        async with aclosing(source) as aiter:
            async for item in aiter:
                if last is token or item != last:
                    last = item
                    await output(item)

class RateLimit(TrioSection):
    """Limits data rate of an input to a certain interval.

    The first item received is transmitted and triggers a timer. Any other items received while
    the timer is active are discarded. After the timer runs out, the cycle can repeat.

    Per subject rate limiting is supported by supplying either a hashable value or a callable as
    subject.
    In case of a hashable value, each received item is assumed to be a mapping, with the subject
    indicating the key containing the value that should be used for rate limiting.
    If a callable is supplied, it will be called with the item as argument and should return a
    hashable value.

    :param interval: Minimum number of seconds between each sent item.
    :type interval: float
    :param source: Input when used as first section.
    :type source: Optional[AsyncIterable[Any]]
    :param subject: Subject for per subject rate limiting.
    :type subject: Optional[]
    """
    def __init__(self,
                 interval,
                 source: Optional[AsyncIterable[Any]] = None,
                 *,
                 subject: Optional[Union[Hashable, Callable[[Any], Hashable]]] = None):
        super().__init__()
        self.source = source
        self.interval = interval
        self.subject = subject

    async def refine(self, input, output):
        if input:
            source = input
        elif self.source:
            source = self.source
        else:
            raise RuntimeError('No input provided.')

        if self.subject is None:
            get_subject = lambda item: None
        elif callable(self.subject):
            get_subject = self.subject
        else:
            get_subject = lambda item: item[self.subject]

        timestamps = {}
        async with aclosing(source) as aiter:
            async for item in aiter:
                now = trio.current_time()
                subject = get_subject(item)
                then = timestamps.get(subject)
                if then is None or now - then > self.interval:
                    timestamps[subject] = now
                    await output(item)



"""The :class:`Pipeline` class is a composable a stream processor. It consists of a chain of
:ref:`pipeline section <Welding sections together>` compatible objects, which each handle a single
stream processing operation.

The stream processing results are accessed by calling :meth:`Pipeline.tap` to create an
output channel. Each pipeline can have multiple open taps, each receiving a copy of the
output stream.

The pipeline can also be extended dynamically with new pipeline sections with
:meth:`Pipeline.extend`, adding additional processing.
"""

import math
from typing import AsyncContextManager, Sequence

import trio
from async_generator import aclosing, asynccontextmanager

from .sections.weld import weld
from ._tap import Tap

class Pipeline:
    """The main Slurry ``Pipeline`` class.

    .. note::
        Do not instantiate a ``Pipeline`` class manually. Use :meth:`create`
        instead. It returns an async context manager which manages the pipeline lifetime.

    Fields:

    * ``sections``: The sequence of pipeline sections contained in the pipeline.
    * ``nursery``: The :class:`trio.Nursery` that is executing the pipeline.

    """
    def __init__(self, *sections: Sequence["PipelineSection"],
                 nursery: trio.Nursery,
                 enabled: trio.Event):
        self.sections = sections
        self.nursery = nursery
        self._enabled = enabled
        self._taps = set()

    @classmethod
    @asynccontextmanager
    async def create(cls, *sections: Sequence["PipelineSection"]) -> AsyncContextManager["Pipeline"]:
        """Creates a new pipeline context and adds the given section sequence to it.

        :param Sequence[PipelineSection] \\*sections: One or more
          :mod:`PipelineSection <slurry.sections.weld>` compatible objects.
        """
        async with trio.open_nursery() as nursery:
            pipeline = cls(*sections, nursery=nursery, enabled=trio.Event())
            nursery.start_soon(pipeline._pump) # pylint: disable=protected-access
            yield pipeline
            nursery.cancel_scope.cancel()

    async def _pump(self):
        """Runs the pipeline."""
        await self._enabled.wait()

        async with trio.open_nursery() as nursery:
            output = weld(nursery, *self.sections)

            # Output to taps
            async with aclosing(output) as aiter:
                async for item in aiter:
                    self._taps = set(filter(lambda tap: not tap.closed, self._taps))
                    if not self._taps:
                        # Hmm.. Debatable. Should closing all taps close the pipeline?
                        break
                    for tap in self._taps:
                        nursery.start_soon(tap.send, item)

        # There is no more output to send. Close the taps.
        for tap in self._taps:
            await tap.send_channel.aclose()

    def tap(self, *,
            max_buffer_size: int = 0,
            timeout: float = math.inf,
            retrys: int = 0,
            start: bool = True) -> trio.MemoryReceiveChannel:
        # pylint: disable=line-too-long
        """Create a new output channel for this pipeline.

        Multiple channels can be opened and will receive a copy of the output data.

        If all open taps are closed, the immidiate upstream section or iterable will be closed as well, and no
        further items can be sent, from that point on.

        .. note::
            The output is sent by reference, so if the output is a mutable type and
            a consumer changes it, other consumers will see the changed output.

        :param max_buffer_size: Although not recommended in general, it is possible to
            set a buffer on the output channel. (default ``0``) See
            `Buffering in channels <https://trio.readthedocs.io/en/stable/reference-core.html#buffering-in-channels>`_
            for further advice.
        :type max_buffer_size: int
        :param timeout: Timeout in seconds when attempting to send an item. (default ``math.inf``)
        :type timeout: float
        :param retrys: Number of times to retry sending, if the initial attempt fails.
            (default ``0``)
        :type retrys: int
        :param start: Start processesing when opening this tap. (default ``True``)
        :type start: bool

        :return: A trio ``MemoryReceiveChannel`` from which pipeline output can be pulled.
        """
        send_channel, receive_channel = trio.open_memory_channel(max_buffer_size)
        self._taps.add(Tap(send_channel, timeout, retrys))
        if start:
            self._enabled.set()
        return receive_channel

    def extend(self, *sections: Sequence["PipelineSection"], start: bool = False) -> "Pipeline":
        """Extend this pipeline into a new pipeline.

        :param Sequence[PipelineSection] \\*sections: One or more pipeline sections.
        :param bool start: Start processing when adding this extension. (default: ``False``)
        """
        pipeline = Pipeline(
            self.tap(start=start),
            sections,
            nursery=self.nursery,
            enabled=self._enabled
        )
        self.nursery.start_soon(pipeline._pump) # pylint: disable=protected-access
        return pipeline



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slurry', 'slurry.environments', 'slurry.sections']

package_data = \
{'': ['*']}

install_requires = \
['trio>=0.17.0,<0.18.0']

setup_kwargs = {
    'name': 'slurry',
    'version': '0.10.3',
    'description': 'An async stream processing microframework',
    'long_description': "======\nSlurry\n======\n\n.. image:: https://readthedocs.org/projects/slurry/badge/?version=latest\n   :target: https://slurry.readthedocs.io/en/latest/?badge=latest\n   :alt: Documentation Status\n\n.. image:: https://img.shields.io/pypi/v/slurry.svg\n   :target: https://pypi.org/project/slurry\n   :alt: Latest PyPi version\n\n.. image:: https://github.com/andersea/slurry/workflows/build/badge.svg\n   :target: https://github.com/andersea/slurry\n   :alt: Build Status\n\nAn async stream processing microframework for Python\n\nIntroduction\n------------\n\nSlurry_ builds on the concepts of structured concurrency and memory channels, originating in\nTrio_, and uses them to create a microframework for processing streaming data.\n\nThe basic building blocks of Slurry includes:\n\n- **Pipelines** - An asynchronous context manager which encapsulates a stream process.\n- **Sections** - The individual processing steps.\n- **Taps** - Output channels for the processed stream.\n- **Extensions** - A way to add more processing steps to an existing pipeline.\n\nSlurry avoids using asynchronous generator functions, in favor of the pull-push programming style\nof memory channels. It can be thought of as an asynchronous version of itertools_ - on steroids!\n\nIncluded in the basic library are a number of basic stream processing building blocks, like\n``Map``, ``Chain``, ``Merge`` and ``Zip``, and it is easy to build your own!\n\nDemonstration\n-------------\nEnough talk! Time to see what's up!\n\n.. code-block:: python\n\n   async with Pipeline.create(\n        Zip(produce_increasing_integers(1, max=3), produce_alphabet(0.9, max=3))\n    ) as pipeline, pipeline.tap() as aiter:\n            results = [item async for item in aiter]\n            assert results == [(0,'a'), (1, 'b'), (2, 'c')]\n\nThe example producers (which are not part of the framework) could look like this:\n\n.. code-block:: python\n\n   async def produce_increasing_integers(interval, *, max=3):\n      for i in range(max):\n         yield i\n         if i == max-1:\n               break\n         await trio.sleep(interval)\n\n   async def produce_alphabet(interval, *, max=3):\n      for i, c in enumerate(string.ascii_lowercase):\n         yield c\n         if i == max - 1:\n               break\n         await trio.sleep(interval)\n\nFurther documentation is available on readthedocs_. Check out the `source code`_ on github__.\n\nInstallation\n------------\nStill here? Wanna try it out yourself? Install from PyPI_::\n\n   pip install slurry\n\nSlurry is tested on Python 3.6 or greater. For now, Slurry is Trio only. AnyIO_ support is not\nruled out in the future.\n\nLicense\n-------\nSlurry is licensed under the `MIT license`_.\n\n.. _Slurry: https://github.com/andersea/slurry\n.. _Trio: https://github.com/python-trio/trio\n.. _itertools: https://docs.python.org/3/library/itertools.html\n.. _PyPI: https://pypi.org/\n.. _readthedocs: https://slurry.readthedocs.io/\n.. _`source code`: https://github.com/andersea/slurry\n__ `source code`_\n.. _AnyIO: https://github.com/agronholm/anyio\n.. _MIT license: https://github.com/andersea/slurry/blob/master/LICENSE\n\n",
    'author': 'Anders Ellenshøj Andersen',
    'author_email': 'andersa@ellenshoej.dk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/andersea/slurry',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""Pipeline sections that produce data streams."""
import trio
from async_generator import aclosing

from ..environments import TrioSection

class Repeat(TrioSection):
    """Yields a single item repeatedly at regular intervals.

    If used as a middle section, the input can be used to set the value that is sent. When
    an input is received, it is sent immidiately, and the internal timer resets.

    :param interval: Delay between each transmission.
    :type interval: float
    :param default: Default item to send. If not supplied, will wait for an input.
    :type default: Any

    :raises RuntimeError: If used as a first section and no default is provided.
    """
    def __init__(self, interval: float, *args, **kwargs):
        self.interval = interval
        self.has_default = False
        self.default = None
        if args:
            self.has_default = True
            self.default = args[0]
        elif 'default' in kwargs:
            self.has_default = True
            self.default = kwargs['default']

    async def refine(self, input, output):
        if input is None and not self.has_default:
            # pylint: disable=line-too-long
            raise RuntimeError('If Repeat is used as first section,  default value must be provided.')

        async with trio.open_nursery() as nursery:

            def start_new_repeater(item):
                cancel_scope = trio.CancelScope()
                async def repeater():
                    with cancel_scope:
                        while True:
                            await trio.sleep(self.interval)
                            await output(item)                
                nursery.start_soon(repeater)
                return cancel_scope

            previous_repeater = None

            if self.has_default:
                await output(self.default)
                previous_repeater = start_new_repeater(self.default)

            if input:
                async with aclosing(input) as aiter:
                    async for item in aiter:
                        if previous_repeater:
                            previous_repeater.cancel()
                        await output(item)
                        previous_repeater = start_new_repeater(item)



"""Pipeline sections with age- and volume-based buffers."""
from collections import deque
import math
from typing import Any, AsyncIterable, Callable, Optional, Sequence

import trio
from async_generator import aclosing

from ..environments import TrioSection

class Window(TrioSection):
    """Window buffer with size and age limits.

    ``Window`` iterates an asynchronous sequence and stores each received item in a
    buffer. Each time another item is received, the buffer is filtered by dumping the oldest items
    first, until the configured window conditions for the buffer size and item age are satisfied.
    After filtering, the whole buffer is output as a tuple, with the oldest item first, and the
    newest item last.

    .. Note::
        Items are added to the right side and are removed from the left side of the buffer.

        All items remain in the buffer, unless they are removed by one of the window
        conditions and any item can be output more than once.

    :param max_size: The maximum buffer size.
    :type max_size: int
    :param source: Input when used as first section.
    :type source: Optional[AsyncIterable[Any]]
    :param max_age: Maximum item age in seconds. (default: unlimited)
    :type max_age: float
    :param min_size: Minimum amount of items in the buffer to trigger an output.
    :type min_size: int
    """
    def __init__(self, max_size: int, source: Optional[AsyncIterable[Any]] = None, *,
                 max_age: float = math.inf,
                 min_size: int = 1):
        super().__init__()
        self.source = source
        self.max_size = max_size
        self.max_age = max_age
        self.min_size = min_size

    async def refine(self, input, output):
        if input:
            source = input
        elif self.source:
            source = self.source
        else:
            raise RuntimeError('No input provided.')

        buf = deque()

        async with aclosing(source) as aiter:
            async for item in aiter:
                now = trio.current_time()
                buf.append((item, now))
                while len(buf) > self.max_size or now - buf[0][1] > self.max_age:
                    buf.popleft()
                if len(buf) >= self.min_size:
                    await output(tuple(i[0] for i in buf))

class Group(TrioSection):
    """Groups received items by time based interval.

    Group awaits an item to arrive from source, adds it to a buffer and sets a timer based on
    the ``interval`` parameter. While the timer is active, additional items received are added
    to the buffer. When the timer runs out, or if the buffer size equals ``max_size``, the buffer
    is sent down the pipeline and a new empty buffer is created.

    .. Note::
        The buffer is not sent at regular intervals. The timer is triggered when an item is
        is received into an empty buffer.

        An output buffer will always contain at least one item.

    The items in the buffer can optionally be mapped over, by supplying a mapper function and be
    reduced to a single value, by supplying a reducer function.

    :param interval: Time in seconds from when an item arrives until the buffer is sent.
    :type interval: float
    :param source: Input when used as first section.
    :type source: Optional[AsyncIterable[Any]]
    :param max_size: Maximum number of items in buffer, which when reached, will cause the buffer
        to be sent.
    :type max_size: int
    :param mapper: Optional mapping function used to transform each received item.
    :type mapper: Optional[Callable[[Any], Any]]
    :param reducer: Optional reducer function used to transform the buffer to a single value.
    :type reducer: Optional[Callable[[Sequence[Any]], Any]]
    """
    def __init__(self, interval: float, source: Optional[AsyncIterable[Any]] = None, *,
                 max_size: int = math.inf,
                 mapper: Optional[Callable[[Any], Any]] = None,
                 reducer: Optional[Callable[[Sequence[Any]], Any]] = None):
        super().__init__()
        self.source = source
        self.interval = interval
        self.max_size = max_size
        self.mapper = mapper
        self.reducer = reducer

    async def refine(self, input, output):
        async with trio.open_nursery() as nursery:
            if input:
                source = input
            elif self.source:
                source = self.source
            else:
                raise RuntimeError('No input provided.')

            send_channel, receive_channel = trio.open_memory_channel(0)
            async def pull_task():
                async with send_channel, aclosing(source) as aiter:
                    async for item in aiter:
                        await send_channel.send(item)
            nursery.start_soon(pull_task)

            while True:
                buffer = []
                try:
                    self._add_item(await receive_channel.receive(), buffer)
                    with trio.move_on_after(self.interval):
                        while True:
                            if len(buffer) == self.max_size:
                                break
                            self._add_item(await receive_channel.receive(), buffer)
                except trio.EndOfChannel:
                    if buffer:
                        await output(self._process_result(buffer))
                    break
                await output(self._process_result(buffer))

    def _add_item(self, item, buffer):
        if self.mapper is not None:
            buffer.append(self.mapper(item))
        else:
            buffer.append(item)

    def _process_result(self, buffer):
        if self.reducer is not None:
            return self.reducer(buffer)
        return tuple(buffer)

class Delay(TrioSection):
    """Delays transmission of each item received by an interval.

    Received items are temporarily stored in an unbounded queue, along with a timestamp, using
    a background task. The foreground task takes items from the queue, and waits until the
    item is older than the given interval and then transmits it.

    :param interval: Number of seconds that each item is delayed.
    :type interval: float
    :param source: Input when used as first section.
    :type source: Optional[AsyncIterable[Any]]
    """
    def __init__(self, interval: float, source: Optional[AsyncIterable[Any]] = None):
        super().__init__()
        self.source = source
        self.interval = interval

    async def refine(self, input, output):
        if input:
            source = input
        elif self.source:
            source = self.source
        else:
            raise RuntimeError('No input provided.')
        buffer_input_channel, buffer_output_channel = trio.open_memory_channel(math.inf)

        async def pull_task():
            async with buffer_input_channel, aclosing(source) as aiter:
                async for item in aiter:
                    await buffer_input_channel.send((item, trio.current_time() + self.interval))

        async with trio.open_nursery() as nursery:
            nursery.start_soon(pull_task)
            async for item, timestamp in buffer_output_channel:
                now = trio.current_time()
                if timestamp > now:
                    await trio.sleep(timestamp - now)
                await output(item)
            nursery.cancel_scope.cancel()



"""The threading module implements a synchronous section that runs in a background thread."""
from typing import Any, AsyncIterable, Awaitable, Callable

import trio

from ..sections.abc import SyncSection


class ThreadSection(SyncSection):
    """ThreadSection defines a section interface which uses a synchronous
    :meth:`refine <slurry.sections.abc.SyncSection.refine>` method.
    """

    async def pump(self,
                   input: AsyncIterable[Any],
                   output: Callable[[Any], Awaitable[None]]):
        """Runs the refine method in a background thread with synchronous input and output
        wrappers, which transparently bridges the input and outputs between the parent
        trio event loop and the sync world.

        .. note::
            Trio has a limit on how many threads can run simultaneously. See the
            `trio documentation <https://trio.readthedocs.io/en/stable/reference-core.html#trio-s-philosophy-about-managing-worker-threads>`_
            for more information.
        """

        def sync_input():
            """Wrapper for turning an async iterable into a blocking generator."""
            if input is None:
                return
            try:
                while True:
                    yield trio.from_thread.run(input.__anext__)
            except StopAsyncIteration:
                pass

        await trio.to_thread.run_sync(self.refine,
                                      sync_input(),
                                      lambda item: trio.from_thread.run(output, item))
